//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Descriptor/VectorDescriptor.h
//! @brief     Defines class VectorDescriptor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DESCRIPTOR_VECTORDESCRIPTOR_H
#define BORNAGAIN_GUI_MODEL_DESCRIPTOR_VECTORDESCRIPTOR_H


#include "GUI/Model/Descriptor/DoubleDescriptor.h"
#include <heinz/Vectors3D.h>

//! Describes properties of a 3D vector, consisting of three double values.
//!
//! By using this class, the underlying data scheme is hidden from the user of the data. This e.g.
//! eases SessionItem migration. The underlying implementation can be a R3 vector, a member array
//! of three doubles, three double members, or any other construction to hold the three values.
//!
//! \a VectorProperty
//! \a DoubleDescriptor
//!
class VectorDescriptor {
public:
    VectorDescriptor() = default;

    VectorDescriptor(const QString& label, const QString& tooltip, double* xVar, double* yVar,
                     double* zVar, const std::variant<QString, Unit>& unit)
        : label(label)
        , tooltip(tooltip)
        , x("X", xVar, unit)
        , y("Y", yVar, unit)
        , z("Z", zVar, unit)
    {
        x.limits = RealLimits::limitless();
        y.limits = RealLimits::limitless();
        z.limits = RealLimits::limitless();
        x.decimals = 3;
        y.decimals = 3;
        z.decimals = 3;
        x.tooltip = tooltip;
        y.tooltip = tooltip;
        z.tooltip = tooltip;
    }

    VectorDescriptor(const QString& label, const QString& tooltip,
                     const std::variant<QString, Unit>& unit)
        : label(label)
        , tooltip(tooltip)
    {
        x.limits = RealLimits::limitless();
        y.limits = RealLimits::limitless();
        z.limits = RealLimits::limitless();
        x.decimals = 3;
        y.decimals = 3;
        z.decimals = 3;
        x.tooltip = tooltip;
        y.tooltip = tooltip;
        z.tooltip = tooltip;
        x.unit = unit;
        y.unit = unit;
        z.unit = unit;
        x.label = "x";
        y.label = "y";
        z.label = "z";
    }

    void init(const QString& _label, const QString& _tooltip, const R3* vec,
              const std::variant<QString, Unit>& _unit)
    {
        label = _label;
        tooltip = _tooltip;
        x.limits = RealLimits::limitless();
        y.limits = RealLimits::limitless();
        z.limits = RealLimits::limitless();
        x.decimals = 3;
        y.decimals = 3;
        z.decimals = 3;
        x.tooltip = tooltip;
        y.tooltip = tooltip;
        z.tooltip = tooltip;
        x.unit = _unit;
        y.unit = _unit;
        z.unit = _unit;
        x.label = "x";
        y.label = "y";
        z.label = "z";
        x.set = [=](double v) { const_cast<R3*>(vec)->setX(v); };
        x.get = [=]() { return vec->x(); };
        x.path = [=] { return uid() + "x"; };
        y.set = [=](double v) { const_cast<R3*>(vec)->setY(v); };
        y.get = [=]() { return vec->y(); };
        y.path = [=] { return uid() + "y"; };
        z.set = [=](double v) { const_cast<R3*>(vec)->setZ(v); };
        z.get = [=]() { return vec->z(); };
        z.path = [=] { return uid() + "z"; };
    }


    QString label;   //!< A label text (short, no trailing colon)
    QString tooltip; //!< Tooltip text
    DoubleDescriptor x;
    DoubleDescriptor y;
    DoubleDescriptor z;
    std::function<QString()> uid; //!< unique id describing  this value. Used e.g. for undo/redo
};

#endif // BORNAGAIN_GUI_MODEL_DESCRIPTOR_VECTORDESCRIPTOR_H
