//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/AxesItems.cpp
//! @brief     Implements various axis items
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Device/AxesItems.h"
#include "Base/Axis/FixedBinAxis.h"
#include "GUI/Support/XML/UtilXML.h"

namespace {

namespace Tag {

const QString IsVisible("IsVisible");
const QString Nbins("Nbins");
const QString MinDeg("MinDeg");
const QString MaxDeg("MaxDeg");
const QString Title("Title");

const QString LockMinMax("LockMinMax");
const QString LogScale("LogScale");

} // namespace Tag

namespace Attrib {

const QString value("value");

}

} // namespace

BasicAxisItem::BasicAxisItem(QObject* parent)
    : QObject(parent)
    , m_visible(true)
    , m_nbins(100)
    , m_minDeg(0)
    , m_maxDeg(-1)
    , m_title("")
{
}

BasicAxisItem::~BasicAxisItem() = default;

int BasicAxisItem::binCount() const
{
    return m_nbins;
}

void BasicAxisItem::setBinCount(size_t value)
{
    m_nbins = int(value);
}


double BasicAxisItem::min() const
{
    return m_minDeg;
}

void BasicAxisItem::setMin(double value)
{
    m_minDeg = value;
    emit axisRangeChanged();
}

double BasicAxisItem::max() const
{
    return m_maxDeg;
}

void BasicAxisItem::setMax(double value)
{
    m_maxDeg = value;
    emit axisRangeChanged();
}

QString BasicAxisItem::title() const
{
    return m_title;
}

void BasicAxisItem::setTitle(const QString& title)
{
    m_title = title;
    emit axisTitleChanged();
}

std::unique_ptr<IAxis> BasicAxisItem::createAxis(double scale) const
{
    return std::make_unique<FixedBinAxis>(title().toStdString(), binCount(), min() * scale,
                                          max() * scale);
}

bool BasicAxisItem::isVisible() const
{
    return m_visible;
}

void BasicAxisItem::setVisible(bool b)
{
    m_visible = b;
    emit axisVisibilityChanged();
}

void BasicAxisItem::writeTo(QXmlStreamWriter* writer) const
{
    // visibility
    writer->writeStartElement(Tag::IsVisible);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_visible);
    writer->writeEndElement();

    // nbins
    writer->writeStartElement(Tag::Nbins);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_nbins);
    writer->writeEndElement();

    // min
    writer->writeStartElement(Tag::MinDeg);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_minDeg);
    writer->writeEndElement();

    // max
    writer->writeStartElement(Tag::MaxDeg);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_maxDeg);
    writer->writeEndElement();

    // title
    writer->writeStartElement(Tag::Title);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_title);
    writer->writeEndElement();
}

void BasicAxisItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement())
        readBasicTags(reader);
}

void BasicAxisItem::readBasicTags(QXmlStreamReader* reader)
{
    // visibility
    if (reader->name() == Tag::IsVisible) {
        GUI::Session::XML::readAttribute(reader, Attrib::value, &m_visible);
        GUI::Session::XML::gotoEndElementOfTag(reader, Tag::IsVisible);

        // nbins
    } else if (reader->name() == Tag::Nbins) {
        GUI::Session::XML::readAttribute(reader, Attrib::value, &m_nbins);
        GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Nbins);

        // min
    } else if (reader->name() == Tag::MinDeg) {
        GUI::Session::XML::readAttribute(reader, Attrib::value, &m_minDeg);
        GUI::Session::XML::gotoEndElementOfTag(reader, Tag::MinDeg);

        // max
    } else if (reader->name() == Tag::MaxDeg) {
        GUI::Session::XML::readAttribute(reader, Attrib::value, &m_maxDeg);
        GUI::Session::XML::gotoEndElementOfTag(reader, Tag::MaxDeg);

        // title
    } else if (reader->name() == Tag::Title) {
        GUI::Session::XML::readAttribute(reader, Attrib::value, &m_title);
        GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Title);

    } else
        reader->skipCurrentElement();
}

// ---------------------------------------------------------------------------------------------- //

AmplitudeAxisItem::AmplitudeAxisItem(QObject* parent)
    : BasicAxisItem(parent)
    , m_lockMinMax(false)
    , m_logScale(true)
{
}

void AmplitudeAxisItem::writeTo(QXmlStreamWriter* writer) const
{
    BasicAxisItem::writeTo(writer);

    // lock min max
    writer->writeStartElement(Tag::LockMinMax);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_lockMinMax);
    writer->writeEndElement();

    // log scale
    writer->writeStartElement(Tag::LogScale);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_logScale);
    writer->writeEndElement();
}

void AmplitudeAxisItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement()) {

        // visibility
        if (reader->name() == Tag::LockMinMax) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_lockMinMax);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::LockMinMax);

            // title
        } else if (reader->name() == Tag::LogScale) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_logScale);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::LogScale);

            // members from base class
        } else {
            BasicAxisItem::readBasicTags(reader);
        }
    }
}

bool AmplitudeAxisItem::isLogScale() const
{
    return m_logScale;
}

void AmplitudeAxisItem::setLogScale(bool value)
{
    m_logScale = value;
    emit logScaleChanged(value);
}

bool AmplitudeAxisItem::isLocked() const
{
    return m_lockMinMax;
}

void AmplitudeAxisItem::setLocked(bool locked)
{
    m_lockMinMax = locked;
}
