//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Job/FitParameterItem.h
//! @brief     Defines class FitParameterItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_JOB_FITPARAMETERITEM_H
#define BORNAGAIN_GUI_MODEL_JOB_FITPARAMETERITEM_H

#include "Base/Types/OwningVector.h"
#include "Fit/Param/AttLimits.h"
#include "GUI/Util/ComboProperty.h"

class FitParameterLinkItem;

//! Holds ComboProperty values, accessible from FitparameterModel.
class FitTypeItem : public QObject {
public:
    explicit FitTypeItem(const ComboProperty& type, QObject* parent = nullptr);

    ComboProperty type() const;
    void setType(const ComboProperty& type);

    void writeTo(QXmlStreamWriter* writer) const;
    void readFrom(QXmlStreamReader* reader);

private:
    ComboProperty m_type;
};

//! Holds a floating-point variable, accessible from FitparameterModel.
class FitDoubleItem : public QObject {
public:
    explicit FitDoubleItem(double value, QObject* parent = nullptr);

    double value() const;
    void setValue(double value);

    RealLimits limits() const;
    void setLimits(const RealLimits& limits);

    int decimals() const;
    void setDecimals(int deciamls);

    virtual void writeTo(QXmlStreamWriter* writer) const;
    virtual void readFrom(QXmlStreamReader* reader);
    void readBasicTags(QXmlStreamReader* reader);

private:
    double m_value;
    int m_decimals;      // no need to save/load
    RealLimits m_limits; // no need to save/load
};

//! Holds floating-point variables which can be enabled or disabled, accessible from
//! FitparameterModel
class FitEditableDoubleItem : public FitDoubleItem {
public:
    explicit FitEditableDoubleItem(double value, bool isEnabled, QObject* parent = nullptr);

    bool isEnabled() const;
    void setEnabled(bool enabled);

    void writeTo(QXmlStreamWriter* writer) const override;
    void readFrom(QXmlStreamReader* reader) override;

private:
    bool m_isEnabled;
};

//! Represents a fit parameter in GUI. Contains links to corresponding ParameterItems in a tuning
//! tree.
class FitParameterItem : public QObject {
public:
    explicit FitParameterItem(QObject* parent = nullptr);

    void initMinMaxValues(const RealLimits& limits);

    AttLimits attLimits() const;

    bool isValid() const;

    QString displayName() const;
    void setDisplayName(QString displayName);

    double startValue() const;
    void setStartValue(double start_value);
    QObject* startValueItem() const;

    double minimum() const;
    void setMinimum(double minimum);
    QObject* minimumItem() const;

    double maximum() const;
    void setMaximum(double maximum);
    QObject* maximumItem() const;

    FitParameterLinkItem* addLink(const QString& title, const QString& link);
    void removeLink(const QString& link);

    QVector<FitParameterLinkItem*> linkItems() const;
    QStringList links() const;

    void setTypeCombo(const ComboProperty& type);
    QString currentType() const;
    QObject* typeItem() const;

    void writeTo(QXmlStreamWriter* writer) const;
    void readFrom(QXmlStreamReader* reader);

private:
    void onTypeChange();
    void setLimitEnabled(FitEditableDoubleItem* propertyItem, bool enabled);
    bool isLimited() const;
    bool isFree() const;
    bool isLowerLimited() const;
    bool isUpperLimited() const;
    bool isFixed() const;

    QString m_displayName;
    std::unique_ptr<FitTypeItem> m_typeItem;
    std::unique_ptr<FitDoubleItem> m_startValueItem;
    std::unique_ptr<FitEditableDoubleItem> m_minItem;
    std::unique_ptr<FitEditableDoubleItem> m_maxItem;
    OwningVector<FitParameterLinkItem> m_links;
};

#endif // BORNAGAIN_GUI_MODEL_JOB_FITPARAMETERITEM_H
