//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Job/JobItem.cpp
//! @brief     Implements class JobItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Job/JobItem.h"
#include "Base/Util/Assert.h"
#include "Device/Detector/SimulationAreaIterator.h"
#include "GUI/Model/CatDevice/InstrumentItemCatalog.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Data/IntensityDataItem.h"
#include "GUI/Model/Data/JobItemUtils.h"
#include "GUI/Model/Data/RealItem.h"
#include "GUI/Model/Data/SpecularDataItem.h"
#include "GUI/Model/Device/Instrument.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Job/FitParameterContainerItem.h"
#include "GUI/Model/Job/FitSuiteItem.h"
#include "GUI/Model/Job/MinimizerItem.h"
#include "GUI/Model/Job/ParameterTreeItems.h"
#include "GUI/Model/Model/SessionXML.h" // writeItemAndChildItems
#include "GUI/Model/Sample/ItemWithMaterial.h"
#include "GUI/Model/Sample/MaterialItems.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Support/IO/ItemFileNameUtils.h"
#include "GUI/Support/XML/Streamer.h"
#include "GUI/Support/XML/UtilXML.h"
#include "GUI/Util/DeserializationException.h"
#include "GUI/Util/Error.h"

#include <stdexcept> // domain_error

namespace {

namespace Tag {

const QString SimulationOptions("SimulationOptions");
const QString ParameterContainer("ParameterContainer");
const QString Sample("Sample");
const QString Instrument("Instrument");
const QString Name("Name");
const QString Identifier("Id");
const QString Comments("Comments");
const QString PresentationType("PresentationType");
const QString Activity("Activity");
const QString Duration("Duration");
const QString BeginTime("Begin");
const QString EndTime("End");
const QString Status("Status");
const QString Progress("Progress");
const QString RealItem("RealItem");
const QString SimulatedData("SimulatedData");
const QString DiffData("DiffData");
const QString FitSuite("FitSuite");

} // namespace Tag

namespace Attrib {

const QString value("value");

}

} // namespace

JobItem::JobItem()
    : JobRealBase(M_TYPE)
    , m_simulationOptionsItem(std::make_unique<SimulationOptionsItem>())
    , m_parameterContainer(std::make_unique<ParameterContainerItem>())
    , m_sampleItem(std::make_unique<MultiLayerItem>())
{
    REFACTORED = true;
}

JobItem::~JobItem() = default;

QString JobItem::identifier() const
{
    return m_identifier;
}

void JobItem::setIdentifier(const QString& identifier)
{
    m_identifier = identifier;
}

QString JobItem::jobName() const
{
    return m_name;
}

void JobItem::setJobName(const QString& name)
{
    m_name = name;
    updateDataFileName();
    emit jobNameChanged(name);
}

JobStatus JobItem::status() const
{
    return m_status;
}

void JobItem::setStatus(JobStatus status)
{
    m_status = status;
    if (status == JobStatus::Failed) {
        if (DataItem* dataItem = simulatedDataItem()) {
            if (dataItem->datafield())
                dataItem->datafield()->setAllTo(0.0);
            emit dataItem->emitDataChanged();
        }
    }
    emit jobStatusChanged(status);
}

bool JobItem::isIdle() const
{
    return status() == JobStatus::Idle;
}

bool JobItem::isRunning() const
{
    return status() == JobStatus::Running;
}

bool JobItem::isCompleted() const
{
    return status() == JobStatus::Completed;
}

bool JobItem::isCanceled() const
{
    return status() == JobStatus::Canceled;
}

bool JobItem::isFailed() const
{
    return status() == JobStatus::Failed;
}

bool JobItem::isFitting() const
{
    return status() == JobStatus::Fitting;
}

bool JobItem::isValidForFitting()
{
    return bool(m_realItem);
}

QDateTime JobItem::beginTime() const
{
    return m_beginTime;
}

void JobItem::setBeginTime(const QDateTime& begin_time)
{
    m_beginTime = begin_time;
    emit jobBeginTimeChanged(begin_time);
}

QDateTime JobItem::endTime() const
{
    return m_endTime;
}

void JobItem::setEndTime(const QDateTime& end_time)
{
    m_endTime = end_time;
    emit jobEndTimeChanged(end_time);
}

std::optional<size_t> JobItem::duration() const
{
    QDateTime begin_time = beginTime();
    QDateTime end_time = endTime();
    if (begin_time.isValid() && end_time.isValid() && begin_time < end_time)
        return begin_time.msecsTo(end_time);
    return std::nullopt;
}

QString JobItem::comments() const
{
    return m_comments;
}

void JobItem::setComments(const QString& comments)
{
    m_comments = comments;
    emit jobCommentsChanged(comments);
}

int JobItem::progress() const
{
    return m_progress;
}

void JobItem::setProgress(int progress)
{
    m_progress = progress;
    emit jobProgressChanged(progress);
}

bool JobItem::runImmediately() const
{
    return simulationOptionsItem().runImmediately();
}

bool JobItem::runInBackground() const
{
    return simulationOptionsItem().runInBackground();
}

QString JobItem::presentationType() const
{
    return m_presentationType;
}

void JobItem::setPresentationType(const QString& type)
{
    m_presentationType = type;
}

QString JobItem::activity() const
{
    return m_activity;
}

void JobItem::setActivity(const QString& activity)
{
    m_activity = activity;
}

bool JobItem::isSpecularJob() const
{
    return instrumentItem()->is<SpecularInstrumentItem>();
}

bool JobItem::isIntensityJob() const
{
    return instrumentItem()->is<GISASInstrumentItem>()
           || instrumentItem()->is<OffspecInstrumentItem>()
           || instrumentItem()->is<DepthProbeInstrumentItem>();
}

MultiLayerItem* JobItem::sampleItem()
{
    return m_sampleItem.get();
}

void JobItem::copySampleIntoJob(const MultiLayerItem* sample)
{
    m_sampleItem->initFrom(sample);
}

InstrumentItem* JobItem::instrumentItem() const
{
    return m_instrument.get();
}

Instrument2DItem* JobItem::instrument2DItem()
{
    auto* result = dynamic_cast<Instrument2DItem*>(instrumentItem());
    ASSERT(result);
    return result;
}

void JobItem::copyInstrumentIntoJob(const InstrumentItem* instrument)
{
    m_instrument.reset(instrument->createCopy());
    m_instrument->setId(QUuid::createUuid().toString());
}

const SimulationOptionsItem& JobItem::simulationOptionsItem() const
{
    return *m_simulationOptionsItem;
}

void JobItem::copySimulationOptionsIntoJob(const SimulationOptionsItem& options)
{
    m_simulationOptionsItem = std::make_unique<SimulationOptionsItem>(options);
}

void JobItem::setResults(const SimulationResult& result)
{
    GUI::Model::JobItemUtils::setResults(simulatedDataItem(), result);
    m_simulatedDataItem->setParentAndModel(this, model());
    updateDataFileName();
}

FitSuiteItem* JobItem::fitSuiteItem()
{
    return m_fitSuiteItem.get();
}

FitSuiteItem* JobItem::createFitSuiteItem()
{
    if (m_fitSuiteItem)
        throw Error("JobItem::createFitSuiteItem() -> Error. Attempt to create "
                    "a second FitSuiteItem.");

    m_fitSuiteItem = std::make_unique<FitSuiteItem>();
    return m_fitSuiteItem.get();
}

void JobItem::createFitContainers()
{
    FitSuiteItem* fitSuiteItem = createFitSuiteItem();

    fitSuiteItem->createFitParametersContainer();
    fitSuiteItem->createMinimizerContainer();
}

ParameterContainerItem* JobItem::parameterContainerItem()
{
    return m_parameterContainer.get();
}

FitParameterContainerItem* JobItem::fitParameterContainerItem()
{
    if (FitSuiteItem* item = fitSuiteItem())
        return item->fitParameterContainerItem();

    return nullptr;
}

DataItem* JobItem::createNewDataItem()
{
    DataItem* dataItem;
    if (isSpecularJob())
        dataItem = new SpecularDataItem();
    else if (isIntensityJob())
        dataItem = new IntensityDataItem();
    else
        ASSERT(0);

    dataItem->setParentAndModel(this, model());
    return dataItem;
}

void JobItem::createSumulatedDataItem()
{
    ASSERT(!simulatedDataItem());
    m_simulatedDataItem.reset(createNewDataItem());
}

IntensityDataItem* JobItem::intensityDataItem()
{
    return dynamic_cast<IntensityDataItem*>(m_simulatedDataItem.get());
}

DataItem* JobItem::simulatedDataItem()
{
    return m_simulatedDataItem.get();
}

void JobItem::createDiffDataItem()
{
    ASSERT(!diffDataItem());
    m_diffDataItem.reset(createNewDataItem());

    // Immediately create coordinate system and axes units
    GUI::Model::JobItemUtils::setDataItemCoords(diffDataItem(), instrumentItem());
    diffDataItem()->updateCoords(instrumentItem());

    if (isSpecularJob())
        dynamic_cast<SpecularDataItem*>(diffDataItem())->setDiffPlotStyle();
}

DataItem* JobItem::diffDataItem()
{
    return m_diffDataItem.get();
}

RealItem* JobItem::createRealItem()
{
    ASSERT(!realItem());
    m_realItem = std::make_unique<RealItem>();
    m_realItem->setParentAndModel(this, model());

    return m_realItem.get();
}

void JobItem::copyRealItemIntoJob(const RealItem* srcRealItem)
{
    createRealItem();
    srcRealItem->copyTo(realItem());

    if (isSpecularJob())
        m_realItem->specularDataItem()->setRealPlotStyle();
}

RealItem* JobItem::realItem()
{
    return m_realItem.get();
}

void JobItem::adjustReaDataToJobInstrument()
{
    // Link RealItem to the JobItem's instrument.
    // (re-)Linking is necessary because of following reason
    // 1) Copying of RealItem from RealModel on board of JobItem requires relink to the copied
    //    instrument
    // 2) During relink all masks (if exists) will be converted to the default units of current
    // detector
    realItem()->linkToInstrument(instrumentItem());

    // Copy masks and ROI from RealItem on board of instrument.
    instrumentItem()->importMasks(realItem()->maskContainerItem());

    // Crop RealItem to the region of interest.
    if (instrumentItem()->is<Instrument2DItem>())
        cropRealData();
}

//! Crops RealItem to the region of interest.
void JobItem::cropRealData()
{
    ASSERT(instrumentItem()->is<Instrument2DItem>());

    // Adjust real data to the size of region of interest
    IntensityDataItem* intensityItem = realItem()->intensityDataItem();

    std::unique_ptr<Datafield> origData(intensityItem->datafield()->clone());

    auto* instrument_item = instrument2DItem();
    GUI::Model::JobItemUtils::createDefaultDetectorMap(intensityItem, instrument_item);

    instrument_item->createInstrument()->detector().iterateOverNonMaskedPoints(
        [&](IDetector::const_iterator it) {
            auto* cropped_data = intensityItem->datafield();
            (*cropped_data)[it.roiIndex()] = (*origData)[it.detectorIndex()];
        });

    intensityItem->updateDataRange();
}

void JobItem::writeTo(QXmlStreamWriter* writer) const
{
    // NOTE: The ordering of the XML elements is important in initialization

    // simulation options
    writer->writeStartElement(Tag::SimulationOptions);
    m_simulationOptionsItem->writeContentTo(writer);
    writer->writeEndElement();

    // instrument
    Streamer(writer).write<InstrumentItemCatalog>(Tag::Instrument, m_instrument.get());

    // sample
    writer->writeStartElement(Tag::Sample);
    m_sampleItem->writeContentTo(writer);
    writer->writeEndElement();

    // parameters
    writer->writeStartElement(Tag::ParameterContainer);
    m_parameterContainer->writeContentTo(writer);
    writer->writeEndElement();

    // data members
    writer->writeStartElement(Tag::Name);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_name);
    writer->writeEndElement();

    // identifier
    writer->writeStartElement(Tag::Identifier);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_identifier);
    writer->writeEndElement();

    // activity
    writer->writeStartElement(Tag::Activity);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_activity);
    writer->writeEndElement();

    // presentation type
    writer->writeStartElement(Tag::PresentationType);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_presentationType);
    writer->writeEndElement();

    // begin time
    writer->writeStartElement(Tag::BeginTime);
    GUI::Session::XML::writeAttribute(writer, Attrib::value,
                                      m_beginTime.toString(Qt::ISODateWithMs));
    writer->writeEndElement();

    // end time
    writer->writeStartElement(Tag::EndTime);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_endTime.toString(Qt::ISODateWithMs));
    writer->writeEndElement();

    // status
    writer->writeStartElement(Tag::Status);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, jobStatusToString(m_status));
    writer->writeEndElement();

    // progress
    writer->writeStartElement(Tag::Progress);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_progress);
    writer->writeEndElement();

    // comments
    writer->writeStartElement(Tag::Comments);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_comments);
    writer->writeEndElement();

    // real item
    if (m_realItem) {
        writer->writeStartElement(Tag::RealItem);
        m_realItem->writeTo(writer);
        writer->writeEndElement();
    }

    // simulated data
    if (m_simulatedDataItem) {
        writer->writeStartElement(Tag::SimulatedData);
        GUI::Session::XML::writeItemAndChildItems(writer, m_simulatedDataItem.get());
        writer->writeEndElement();
    }

    // diff data
    if (m_diffDataItem) {
        writer->writeStartElement(Tag::DiffData);
        GUI::Session::XML::writeItemAndChildItems(writer, m_diffDataItem.get());
        writer->writeEndElement();
    }

    // fit suite
    if (m_fitSuiteItem) {
        writer->writeStartElement(Tag::FitSuite);
        m_fitSuiteItem->writeTo(writer);
        writer->writeEndElement();
    }
}

template <class T>
inline T* readItemToEnd(QXmlStreamReader* reader, SessionItem* parent_item, const QString& tag)
{
    T* item = dynamic_cast<T*>(GUI::Session::XML::readItemAndChildItems(reader, parent_item, tag));
    GUI::Session::XML::gotoEndElementOfTag(reader, tag);
    return item;
}

void JobItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement()) {

        // simulation options
        if (reader->name() == Tag::SimulationOptions) {
            m_simulationOptionsItem->readContentFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::SimulationOptions);

            // instrument
        } else if (reader->name() == Tag::Instrument) {
            InstrumentItem* p = nullptr;
            Streamer(reader).read<InstrumentItemCatalog>(Tag::Instrument, p);
            m_instrument.reset(p);

            // parameters
        } else if (reader->name() == Tag::ParameterContainer) {
            m_parameterContainer->readContentFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::ParameterContainer);

            // sample
        } else if (reader->name() == Tag::Sample) {
            m_sampleItem->readContentFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Sample);

            // activity
        } else if (reader->name() == Tag::Activity) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_activity);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Activity);

            // presentation type
        } else if (reader->name() == Tag::PresentationType) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_presentationType);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::PresentationType);

            // comments
        } else if (reader->name() == Tag::Comments) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_comments);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Comments);

            // status
        } else if (reader->name() == Tag::Status) {
            QString status_str;
            GUI::Session::XML::readAttribute(reader, Attrib::value, &status_str);
            m_status = jobStatusFromString(status_str);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Status);

            // progress
        } else if (reader->name() == Tag::Progress) {
            QString progress_str;
            GUI::Session::XML::readAttribute(reader, Attrib::value, &progress_str);
            m_progress = progress_str.toInt();
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Progress);

            // begin time
        } else if (reader->name() == Tag::BeginTime) {
            QString begin_t_str;
            GUI::Session::XML::readAttribute(reader, Attrib::value, &begin_t_str);
            m_beginTime = QDateTime::fromString(begin_t_str, Qt::ISODateWithMs);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::BeginTime);

            // end time
        } else if (reader->name() == Tag::EndTime) {
            QString end_t_str;
            GUI::Session::XML::readAttribute(reader, Attrib::value, &end_t_str);
            m_endTime = QDateTime::fromString(end_t_str, Qt::ISODateWithMs);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::EndTime);

            // identifier
        } else if (reader->name() == Tag::Identifier) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_identifier);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Identifier);

            // name
        } else if (reader->name() == Tag::Name) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_name);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Name);

            // real item
        } else if (reader->name() == Tag::RealItem) {
            createRealItem()->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::RealItem);

            // simulated data
        } else if (reader->name() == Tag::SimulatedData) {
            m_simulatedDataItem.reset(readItemToEnd<DataItem>(reader, this, Tag::SimulatedData));

            // diff data
        } else if (reader->name() == Tag::DiffData) {
            m_diffDataItem.reset(readItemToEnd<DataItem>(reader, this, Tag::DiffData));

            // fit suite
        } else if (reader->name() == Tag::FitSuite) {
            createFitSuiteItem()->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::FitSuite);

        } else
            reader->skipCurrentElement();
    }
}

//! Updates the name of file to store intensity data.

void JobItem::updateDataFileName()
{
    if (DataItem* item = simulatedDataItem())
        item->setFileName(GUI::Model::FilenameUtils::jobResultsFileName(jobName()));

    if (RealItem* real = realItem()) {
        if (DataItem* item = real->dataItem())
            item->setFileName(GUI::Model::FilenameUtils::jobReferenceFileName(jobName()));

        if (DataItem* item = real->nativeData())
            item->setFileName(GUI::Model::FilenameUtils::jobNativeDataFileName(identifier()));
    }
}
