//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/ParticleItem.cpp
//! @brief     Implements class ParticleItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/CatSample/FormFactorItemCatalog.h"
#include "GUI/Model/CatSample/RotationItemCatalog.h"
#include "GUI/Model/Sample/CoreAndShellItem.h"
#include "GUI/Model/Sample/FormFactorItems.h"
#include "GUI/Model/Sample/MaterialItem.h"
#include "GUI/Support/XML/Serialize.h"
#include "Sample/Particle/IFormFactor.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Scattering/Rotations.h"

namespace {

const QString abundance_tooltip = "Proportion of this type of particles normalized to the \n"
                                  "total number of particles in the layout";

const QString position_tooltip = "Relative position of the particle's reference point \n"
                                 "in the coordinate system of the parent (nm)";

} // namespace

ParticleItem::ParticleItem(const MaterialItems* materials)
    : ItemWithMaterial(materials)
    , ItemWithParticles(abundance_tooltip, position_tooltip)
{
    m_formFactor.init<FormFactorItemCatalog>("Form Factor", "", "formFactor");
    m_formFactor.set(new CylinderItem());
}

void ParticleItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwValue(s, "MaterialIdentifier", m_materialIdentifier);
    Serialize::rwProperty(s, m_abundance);
    Serialize::rwProperty(s, m_position);
    Serialize::rwSelected<RotationItemCatalog>(s, m_rotation);
    Serialize::rwSelected<FormFactorItemCatalog>(s, m_formFactor);
}

std::unique_ptr<Particle> ParticleItem::createParticle() const
{
    auto domainMaterial = materialItem()->createMaterial();

    auto particle = std::make_unique<Particle>(*domainMaterial, *m_formFactor->createFormFactor());
    particle->setAbundance(abundance());
    particle->setParticlePosition(position());
    if (auto r = createRotation(); r && !r->isIdentity())
        particle->setRotation(*r);

    return particle;
}

void ParticleItem::setFormFactor(FormFactorItem* p)
{
    m_formFactor.set(p);
}

FormFactorItem* ParticleItem::formfactor() const
{
    return m_formFactor.get();
}

QVector<ItemWithParticles*> ParticleItem::containedItemsWithParticles() const
{
    return {};
}
