//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Common/DoubleSpinBox.cpp
//! @brief     Implements class DoubleSpinBox
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Common/DoubleSpinBox.h"
#include "GUI/View/Tool/EditUtil.h"
#include <QWheelEvent>

DoubleSpinBox::DoubleSpinBox(const DoubleDescriptor& d, bool easyScrollable, int precision,
                             double stepSize, QWidget* parent)
    : QDoubleSpinBox(parent)
    , m_valueDescriptor(d)
    , easyScrollable(easyScrollable)
    , m_precision(precision)
{
    setFocusPolicy(Qt::StrongFocus);
    GUI::View::EditUtil::configSpinbox(this, d.decimals, d.limits);
    setToolTip(d.tooltip);
    setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);

    if (std::holds_alternative<QString>(m_valueDescriptor.unit))
        setDisplayUnit(Unit::other);
    else
        setDisplayUnit(std::get<Unit>(m_valueDescriptor.unit));

    QObject::connect(this, QOverload<double>::of(&QDoubleSpinBox::valueChanged), this,
                     &DoubleSpinBox::onDisplayValueChanged);
    setSingleStep(stepSize);
}
void DoubleSpinBox::setDisplayUnit(Unit displayUnit)
{
    m_displayUnit = displayUnit;

    if (m_showUnitAsSuffix) {
        const QString suffix = displayUnitAsString();
        if (suffix.isEmpty())
            setSuffix("");
        else
            setSuffix(" " + suffix);
    }

    QSignalBlocker b(this);
    setValue(toDisplayValue(m_valueDescriptor.get()));
}

double DoubleSpinBox::toDisplayValue(double baseValue) const
{
    return convert(baseValue, baseUnit(), m_displayUnit);
}

double DoubleSpinBox::toBaseValue(double displayValue) const
{
    return convert(displayValue, m_displayUnit, baseUnit());
}

QString DoubleSpinBox::displayUnitAsString() const
{
    if (std::holds_alternative<QString>(m_valueDescriptor.unit))
        return std::get<QString>(m_valueDescriptor.unit);

    return unitAsString(m_displayUnit);
}

const DoubleDescriptor& DoubleSpinBox::valueDescriptor() const
{
    return m_valueDescriptor;
}

void DoubleSpinBox::setBaseValue(double baseValue)
{
    setValue(toDisplayValue(baseValue));
}

void DoubleSpinBox::wheelEvent(QWheelEvent* event)
{
    if (hasFocus() || easyScrollable)
        QDoubleSpinBox::wheelEvent(event);
    else
        event->ignore();
}

void DoubleSpinBox::onDisplayValueChanged(double newDisplayValue)
{
    emit baseValueChanged(toBaseValue(newDisplayValue));
}

Unit DoubleSpinBox::baseUnit() const
{
    if (std::holds_alternative<QString>(m_valueDescriptor.unit))
        return Unit::other;

    return std::get<Unit>(m_valueDescriptor.unit);
}

void DoubleSpinBox::updateValue()
{
    QSignalBlocker b(this);
    setBaseValue(m_valueDescriptor.get());
}

QString DoubleSpinBox::textFromValue(double val) const
{
    return QString::number(val, 'f', m_precision);
}
