//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Common/DoubleSpinBox.h
//! @brief     Defines class DoubleSpinBox
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_VIEW_COMMON_DOUBLESPINBOX_H
#define BORNAGAIN_GUI_VIEW_COMMON_DOUBLESPINBOX_H
#include "GUI/Model/Descriptor/DoubleDescriptor.h"
#include <QDoubleSpinBox>

//! SpinBox for DoubleDescriptors, supporting units.
class DoubleSpinBox : public QDoubleSpinBox {
    Q_OBJECT
public:
    //! Create a DoubleSpinBox with the information found in a DoubleDescriptor.
    //!
    //! The spin box will be fully initialized (tooltip, limits, unit, current value, size policy).
    //! Furthermore, the spin box will prohibit accidental changes by the mouse wheel. Otherwise it
    //! would be dangerous if the spin box is on a scrollable form - unintended and unnoticed
    //! changes would take place when just scrolling through the form.
    DoubleSpinBox(const DoubleDescriptor& d, bool easyScrollable = false, int precision = 2,
                  double singleStepSize = 0.1, QWidget* parent = nullptr);
    //! Set a display unit.
    //!
    //! The caller has to make sure that the new display unit has a conversion to/from the contained
    //! base value unit.
    void setDisplayUnit(Unit displayUnit);

    //! Set the base value (unit is the one of the contained descriptor).
    void setBaseValue(double baseValue);
    //! The display unit as human readable string.
    QString displayUnitAsString() const;

    //! The descriptor on which this spinbox operates.
    const DoubleDescriptor& valueDescriptor() const;

    //! Returns the unit of the contained DoubleDescriptor.
    //!
    //! If the unit is defined as a string, this method returns Unit::other. To get the string, use
    //! valueDescriptor().unit
    Unit baseUnit() const;

    //! Update the shown value to the one contained in the value descriptor.
    //!
    //! No signal will be emitted if the new value has changed.
    void updateValue();


signals:
    //! Emitted whenever the value changes.
    //!
    //! newBaseValue is in the unit of the valueDescriptor.
    void baseValueChanged(double newBaseValue);

protected:
    void wheelEvent(QWheelEvent* event) override;

private:
    void onDisplayValueChanged(double newDisplayValue);

    double toDisplayValue(double baseValue) const;
    double toBaseValue(double displayValue) const;

    using QDoubleSpinBox::setValue; // To hide from usage

private:
    Unit m_displayUnit = Unit::unitless;

    DoubleDescriptor m_valueDescriptor;
    bool easyScrollable;

    //! it was decided to not show the unit as a suffix. However, this may be user
    //! selectable once, therefore the code is kept and controlled by this flag
    bool m_showUnitAsSuffix = false;

    //! stores the decimal places used for displaying the value. A user can still
    //! type more decimal places, these places will get rounded up to the places allowed.
    const int m_precision;

    // QDoubleSpinBox interface
public:
    QString textFromValue(double val) const override;
};


#endif // BORNAGAIN_GUI_VIEW_COMMON_DOUBLESPINBOX_H
